from . import ParadeCommand
from ..core.engine import Engine
from ..core.task import Task
from ..utils.log import logger


class ExecCommand(ParadeCommand):
    """
    The exec command to run a flow or a set tasks,
    if the tasks to execute have dependencies on each other,
    parade will handle them correctly
    """
    requires_workspace = True

    def run_internal(self, context, **kwargs):
        engine = Engine(context)

        tasks = kwargs.get('task')
        force = kwargs.get('force')
        no_dag = kwargs.get('no_dag')
        logger.debug('prepare to execute tasks {}'.format(tasks))

        if len(tasks) == 0:
            tasks = list(context.list_tasks())
            logger.info('no task provided, use detected {} tasks in workspace {}'.format(len(tasks), tasks))

        if len(tasks) == 1:
            logger.info('single task {} provided, ignore its dependencies'.format(tasks[0]))
            no_dag = True

        if no_dag:
            for task in tasks:
                retcode, _, _ = engine.execute(task, force=force)
                if retcode != Task.RET_CODE_SUCCESS:
                    return retcode
        else:
            engine.execute_dag(*tasks)

        return Task.RET_CODE_SUCCESS

    def short_desc(self):
        return 'execute a flow or a set of tasks'

    def config_parser(self, parser):
        parser.add_argument('--no-dag', action="store_true", help='execute tasks without considering dependencies')
        parser.add_argument('--force', action="store_true", help='force the task to execute')
        parser.add_argument('task', nargs='*', help='the task to schedule')



# from parade.connection import Connection
# import pandas as pd
# import os
# import datetime
#
#
# class LocalFile(Connection):
#     def store(self, df, table, **kwargs):
#         target_path = self.datasource.db if self.datasource.db else '.'
#         export_type = self.datasource.protocol if self.datasource.protocol else 'excel'
#         if_exists = kwargs.get('if_exists', 'replace')
#
#         os.makedirs(target_path, exist_ok=True)
#
#         if export_type == 'excel':
#             target_file = os.path.join(target_path, table + '-' + str(datetime.date.today())) + '.xlsx'
#             if if_exists == 'replace' and os.path.exists(target_file):
#                 os.remove(target_file)
#             writer = pd.ExcelWriter(target_file, engine='xlsxwriter')
#             df.to_excel(writer, index=False)
#             writer.save()
#         elif export_type == 'csv':
#             target_file = os.path.join(target_path, table + '-' + str(datetime.date.today())) + '.csv'
#             if if_exists == 'replace' and os.path.exists(target_file):
#                 os.remove(target_file)
#             df.to_csv(target_file, index=False, chunksize=4096)
#         elif export_type == 'pickle':
#             target_file = os.path.join(target_path, table + '-' + str(datetime.date.today())) + '.dat'
#             if if_exists == 'replace' and os.path.exists(target_file):
#                 os.remove(target_file)
#             df.to_pickle(target_file)
#         else:
#             raise NotImplementedError("export type {} is not supported".format(export_type))
#
#     def load_query(self, query, **kwargs):
#         raise NotImplementedError
#
#     def load(self, table, **kwargs):
#         raise NotImplementedError
#
#     def init_record_if_absent(self):
#         raise NotImplementedError
#
#     @property
#     def accept(self):
#         raise NotImplementedError
#
#     @property
#     def can_produce(self):
#         raise NotImplementedError
#
#     def last_record(self, task_name):
#         raise NotImplementedError
#
#     def create_record(self, task_name, new_checkpoint):
#         raise NotImplementedError
#
#     def commit_record(self, txn_id):
#         raise NotImplementedError
#
#     def rollback_record(self, txn_id, err):
#         raise NotImplementedError



from . import ParadeCommand
from ..dagstore import DAGStore
from ..utils.modutils import get_class
from ..utils.log import logger


class MakeDAGCommand(ParadeCommand):

    requires_workspace = True

    def run_internal(self, context, **kwargs):
        config = context.conf
        dagstore_driver = config['dagstore.driver']
        dagstore_cls = get_class(dagstore_driver, DAGStore, 'parade.dagstore', context.name + ".contrib.dagstore")
        dagstore = dagstore_cls(context)

        tasks = kwargs.get('task')
        dag_key = kwargs.get('flow_name')
        if not dag_key:
            dag_key = context.name
        cron = kwargs.get('cron')

        if len(tasks) == 0:
            tasks = context.list_tasks()
            logger.info('no task provided, use detected {} tasks in workspace {}'.format(len(tasks), tasks))

        dagstore.create(*tasks, dag_key=dag_key, cron=cron)

    def short_desc(self):
        return 'create a dag (flow) with a set of tasks'

    def config_parser(self, parser):
        parser.add_argument('--cron', dest='cron', help='set the schedule settings')
        parser.add_argument('--dag-key', dest='flow_name', help='set the key for the dag')
        parser.add_argument('task', nargs='*', help='the task to compose the dag')




import datetime

import pandas as pd
from sqlalchemy import create_engine, MetaData, Column, types, Index, Table
from sqlalchemy.sql import functions

from parade.utils.log import logger
from . import Connection, Datasource


class RDBConnection(Connection):
    def __init__(self, datasource):
        Connection.__init__(self, datasource)
        assert isinstance(datasource, Datasource), 'Invalid connection provided'
        # assert connection.host is not None, 'host of connection is required'
        # assert connection.port is not None, 'port of connection is required'
        # assert connection.db is not None, 'db of connection is required'
        # assert connection.protocol is not None, 'protocol of connection is required'
        self.metadata = MetaData()
        self.task_table = Table('sys_task_record', self.metadata,
                                Column('id', types.BigInteger, autoincrement=True, primary_key=True),
                                Column('task', types.String(64), nullable=False),
                                Column('checkpoint', types.DateTime, default=datetime.datetime.now),
                                Column('create_time', types.DateTime, default=datetime.datetime.now),
                                Column('commit_time', types.DateTime, default=datetime.datetime.now),
                                Column('update_time', types.DateTime, default=datetime.datetime.now,
                                       onupdate=datetime.datetime.now),
                                Column('status', types.Integer, default=0),
                                Column('message', types.String(256), default='OK'),

                                Index('idx_task_create', 'task', 'create_time'),
                                )

    def open(self):
        uri = self.datasource.uri
        if uri is None:
            authen = None
            uripart = self.datasource.host + ':' + str(self.datasource.port) + '/' + self.datasource.db
            if self.datasource.user is not None:
                authen = self.datasource.user
            if authen is not None and self.datasource.password is not None:
                authen += ':' + self.datasource.password
            if authen is not None:
                uripart = authen + '@' + uripart
            uri = self.datasource.protocol + '://' + uripart
        pandas_conn = create_engine(uri, encoding="utf-8")
        return pandas_conn

    def load(self, table, **kwargs):
        return self.load_query('select * from {}'.format(table))

    def load_query(self, query, **kwargs):
        conn = self.open()
        return pd.read_sql_query(query, con=conn)

    def store(self, df, table, **kwargs):
        assert isinstance(df, pd.DataFrame), "Invalid data type"
        if_exists = kwargs.get('if_exists', 'fail')
        chunksize = kwargs.get('chunksize', None)
        pkey = kwargs.get('pkey', None)
        indexes = kwargs.get('indexes', [])
        checkpoint_column = kwargs.get('checkpoint_column', None)
        checkpoint = kwargs.get('checkpoint')
        last_checkpoint = kwargs.get('last_checkpoint')

        _conn = self.open()

        if if_exists == 'append' or if_exists == 'update':
            target_table = Table(table, MetaData(), autoload=True, autoload_with=_conn)
            assert checkpoint_column is not None, "checkpoint_column is required in update mode!"
            assert (isinstance(checkpoint_column, tuple) and len(checkpoint_column) == 2) or isinstance(checkpoint_column, str), "checkpoint_column can only be str or 2-tuple!"

            if isinstance(checkpoint_column, tuple):
                (create_time_column, update_time_column) = checkpoint_column
            else:
                create_time_column = checkpoint_column
                update_time_column = checkpoint_column

            # delete extra records over last checkpoint in append/update mode
            clear_ins = target_table.delete().where(Column(update_time_column) >= last_checkpoint)
            _conn.execute(clear_ins)

            if if_exists == 'update':
                assert pkey is not None, "primary key is required in update mode!"
                assert isinstance(pkey, str), "update mode only support single primary key"
                update_df = df[df[create_time_column] < last_checkpoint]
                if not update_df.empty:
                    logger.info("find {} records to update".format(len(update_df)))
                    update_keys = list(update_df[pkey])
                    delete_ins = target_table.delete().where(Column(pkey).in_(update_keys))
                    _conn.execute(delete_ins)
                if_exists = 'append'

        schema = None
        if table.find('.') >= 0:
            toks = table.split('.', 1)
            schema = toks[0]
            table = toks[1]

        float_columns = list(df.select_dtypes(include=['float64', 'float']).keys())
        if len(float_columns) > 0:
            logger.warn(
                    "Detect columns with float types {}, you better check if this is caused by NAN-integer column issue of pandas!".format(
                            list(float_columns)))

        typehints = dict()
        obj_columns = list(df.select_dtypes(include=['object']).keys())

        if len(obj_columns) > 0:
            logger.warn(
                    "Detect columns with object types {}, which is automatically converted to *VARCHAR(256)*, you can override this by specifying type hints!".format(
                            list(obj_columns)))
        import sqlalchemy.types as sqltypes
        typehints.update(dict((k, sqltypes.VARCHAR(256)) for k in obj_columns))

        # TODO: upddate typehints with user-specified one
        _typehints = kwargs.get('typehints', {})
        from parade.type import stdtype_to_sqltype
        for col, stdtype in _typehints.items():
            logger.info("Column [{}] is set to type [{}]".format(col, str(stdtype)))
            typehints[col] = stdtype_to_sqltype(stdtype)

        def _chunks(_df, _chunksize):
            """Yield successive n-sized chunks from l."""
            for i in range(0, len(_df), _chunksize):
                yield df[i:i + _chunksize]

        # still write to database for empty dataframe
        if df.empty:
            df.to_sql(name=table, con=_conn, index=False, schema=schema, if_exists=if_exists, dtype=typehints)
            logger.warn("Write to {}: empty dataframe".format(table))
        else:
            for idx, chunk in enumerate(_chunks(df, chunksize)):
                if_exists_ = 'append' if idx > 0 else if_exists
                chunk.to_sql(name=table, con=_conn, index=False, schema=schema, if_exists=if_exists_, dtype=typehints)
                logger.info("Write to {}: rows #{}-#{}".format(table, idx * chunksize, (idx + 1) * chunksize))

        if if_exists == 'replace':
            if pkey:
                pkeys = pkey if isinstance(pkey, str) else ','.join(pkey)
                _conn.execute('ALTER TABLE {} ADD PRIMARY KEY ({})'.format(table, pkeys))

            for index in indexes:
                index_str = index if isinstance(index, str) else ','.join(index)
                index_name = index if isinstance(index, str) else '_'.join(index)
                _conn.execute('ALTER TABLE {} ADD INDEX idx_{} ({})'.format(table, index_name, index_str))

    def init_record_if_absent(self):
        _conn = self.open()
        if not self.task_table.exists(_conn):
            try:
                self.task_table.create(_conn)
            except:
                pass

    def last_record(self, task_name):
        _conn = self.open()
        _query = self.task_table.select(). \
            where(self.task_table.c.task == task_name). \
            where(self.task_table.c.status == 1). \
            order_by(self.task_table.c.create_time.desc()).limit(1)
        _last_record = _conn.execute(_query).fetchone()

        if _last_record is not None:
            return dict(_last_record)
        return None

    def create_record(self, task_name, new_checkpoint):
        _conn = self.open()

        # 创建待提交checkpoint
        ins = self.task_table.insert().values(task=task_name, checkpoint=new_checkpoint)
        return _conn.execute(ins).inserted_primary_key[0]

    def commit_record(self, txn_id):
        _conn = self.open()
        sql = self.task_table.update(). \
            where(self.task_table.c.id == txn_id). \
            values(status=1, commit_time=functions.now())
        _conn.execute(sql)

    def rollback_record(self, txn_id, err):
        _conn = self.open()
        sql = self.task_table.update(). \
            where(self.task_table.c.id == txn_id). \
            values(status=2, message=str(err))
        _conn.execute(sql)



from . import DAGStore
from ..utils.log import logger


class ParadeDAGStore(DAGStore):
    """
    This is *JUST A PLACEHOLDER* implementation
    """

    def create(self, *tasks, cron=None, dag_key=None):
        logger.warn('This is JUST-A-PLACEHOLDER, no dag is created')
        return None

    def delete(self, flow_name):
        pass

